/*
 *	celcoo:  celestial coordinate conversion
 *
 *	This program converts right ascension and declination to hour angle,
 *	altitude, and azimuth, for a given place and time.
 *
 *	Type "celcoo" with no parameters to get the usage message.
 */

#define	VERNUM	"2.0"

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
long time();

/* define bool symbols */
typedef enum {
	false = 0,
	true = 1
} bool;

#define PI	3.141592653589793

/* find number of elements in an array */
#define NUMELEM(array)  (sizeof (array) / sizeof (array[0]))

double aint();
double pfmod();
void praltaz();
double rad;
int hms_str2float(char *str, float *ans_ptr);
int dms_str2float(char *str, float *ans_ptr);
void float_to_3_longs(float f, long *a_ptr, long *b_ptr, long *c_ptr);
void error(char *format, ...);


long th;
long tm;
long ts;
long mo;				/* month num Jan = 1	*/
long day;			/* day of month		*/
long yr;				/* four digits		*/
long tz=6;			/* Default time zone	*/
char *tzs = "(CST)";		/* Default time zone string */
char *dtzs = "(CDT)";		/* Default daylight savings time string */
double greensid;		/* Greenwich sideral time in radians */
long latd, latm, lats;		/* north latitude degrees, minutes, seconds */
long lond, lonm, lons;		/* west longitude degrees, minutes, seconds */
long ra_h, ra_m, ra_s;		/* right ascension hours, minutes, seconds */
long dec_d, dec_m, dec_s;	/* declination degrees, minutes, seconds */
double latitude;		/* north latitude radians */
double longitude;		/* west longitude radians */
char	*mtab[] = {
	"dummy", "January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December",
	};
char *Prog;



main(argc, argv)
int	argc;
char	*argv[];
{
	long mm,dd,yy, aa1,bb1;
	double otime;	/* hundredths of an hour since start of Julian day */
	double jd,year,month,oday;

	double l,a,e,i,w1,w2,M,M0,M1,M2,M4,M5,M6,M7,M8;
	double a0,a1,a2,a3,RA,DEC;
	double ECC,nu,r,u,ll,b,lonpert,radpert;
	double esun,Lsun,Cen,Stheta,Snu,Sr;
	double N,D,epli,thapp,omeg;
	double nu2,P,Q,S,V,W,ze,l1pert,epert,w1pert,apert;
	double psi,H,G,eta,oth;
	double ra, dec;
	double ra_rad, dec_rad;	/* in radians */
	
	char sign;
	time_t sec_1970;
	long h, m;
	struct tm *pt;

	Prog = argv[0];
	rad = PI/180.0;

	/* get system date & time, default if user doesn't supply them */

	sec_1970 = time((time_t *)0);  
	pt = localtime(&sec_1970);  

	th = pt->tm_hour;
	tm = pt->tm_min;
	ts = pt->tm_sec;
	yr = pt->tm_year + 1900;	/* unix uses last two digits	*/
	mo = pt->tm_mon + 1;		/* unix uses Jan = 0	*/
	day = pt->tm_mday;

	/* convert tz to daylight savings time */
	if (pt->tm_isdst) {
		tz--;
		tzs = dtzs;	
	}

	/* set default location to be Elgin, Illinois */
	latd = 42, latm =  2; lats = 0;
	lond = 88, lonm = 16; lons = 0;

	/* get user's arguments to override defaults */
	initopts(argc,argv);

	latitude  = ((double)latd + (double)latm/60 + (double)lats/3600) * rad;
	longitude = ((double)lond + (double)lonm/60 + (double)lons/3600) * rad;

	printf("celcoo	Version %s \n", VERNUM);

	/* copy my time vars */
	mm = mo;
	dd = day;
	yy = yr;
	otime = ((double)th + (double)tm/60.0 + (double)ts/3600.0) * 100.0;
	/* tz is ok */
	/* show the month name instead of just the numbers */
	if ( (mm < 1) || ( mm > 12) ) {
		printf("Invalid month (%ld) \n", mm);
		usage();
		exit(1);
	}

	printf(" Date:  %s %ld, %ld", mtab[(int)mm], dd, yy);
	printf("   Time: %02ld:%02ld:%02ld %s", th, tm, ts, tzs);

	/*  Calculate the Julian date for the time of observation */
	
	oday = ((double)dd + (otime +(double)tz*100.0)/2400.0);
	if (mm > 2)
	{
		year = yy;
		month = mm;
	}
	else
	{
		year = yy - 1;
		month = mm + 12;
	}
	aa1 = (long) (year/100);
	bb1 = 2 - aa1 + (long)(aa1/4);
	jd = aint(365.25*year) + aint(30.6001*(month + 1));
	jd = jd  + oday + 1720994.5;
	if (year < 0)
	{
		error("negative year not allowed");
	}
	if((year + month/100) > 1582.10)
	{
		jd = jd + bb1;
	}
	printf("   Julian date: %f\n", jd);

	if (latd < 0 || latm < 0 || lats < 0) {
		printf(" Place: %ldo %ld' %ld\" south latitude, ",
				-latd, -latm, -lats);
	} else {
		printf(" Place: %ldo %ld' %ld\" north latitude, ",
				latd, latm, lats);
	}
	if (lond < 0 || lonm < 0 || lons < 0) {
		printf("%ldo %ld' %ld\" east longitude\n", -lond, -lonm, -lons);
	} else {
		printf("%ldo %ld' %ld\" west longitude\n", lond, lonm, lons);
	}
	
/* calculate the Greenwich siderial time */
	/*
	 * Find the Greenwich sideral time, using "JD2000" as a reference.  See
	 * http://www.pietro.org/Astro_Util_StaticDemo/FDetailSideralConv.html
	 * or http://www.astro.washington.edu/deutsch-bin/getpro/
	 * library32.html?C%2LST for info on this.  Convert to radians.
	 */
	greensid = 280.46061837 + (jd - 2451545.0) * 360.98564736629;
	greensid /= 360.0;
	greensid = pfmod(greensid, 1.);
	greensid *= 2 * PI;

	ra = ra_h * 15.0 + ra_m / 4.0 + ra_s / 240.0;
	dec = dec_d + dec_m / 60.0 + dec_s / 3600.0;
	ra_rad = ra / 180 * PI;
	dec_rad = dec / 180 * PI;

	sign = dec >= 0 ? '+' : '-';
	printf("right ascension = %2ldh %2ldm %2lds    declination = %c%2ldo %2ld' %2ld\"\n",
		ra_h, ra_m, ra_s, sign, labs(dec_d), labs(dec_m), labs(dec_s));

	praltaz(ra_rad, dec_rad);	/* print altitude and azimuth */

	putchar('\n');
	exit(0);
} /* end of program main */


initopts(argc,argv)
int argc;
char *argv[];
{
	bool got[256];		/* guard against multiple usages of options */
	char timebuf[20];	/* copy of -t argument */
	int idx;		/* loop index */
	int c;			/* an option character */
	char *c_ptr;
	float temp;


	/* init that no options have been parsed yet */
	for (idx = 0; idx < NUMELEM(got); idx++) {
		got[idx] = false;
	}

	/*
	 * Loop through the options, storing that information.  Make sure none
	 * are used multiple times.
	 */
	while ((c = getopt(argc, argv, "R:D:d:t:z:a:o:")) != -1) {
		if (got[c]) {
			error("option -%c is used multiple times", c);
		}
		got[c] = true;

		switch (c) {
		case 'd':
			if (sscanf(optarg, "%ld/%ld/%ld", &mo, &day, &yr) != 3){
				error("invalid format %s for -d option",
						optarg);
			}
			if (mo < 1 || mo > 12) {
				error("invalid month %ld in -d option", mo);
			}
			if (day < 1 || day > 31) {
				error("invalid day %ld in -d option", day);
			}
			if (yr < 1) {
				error("invalid year %ld in -d option", yr);
			}
			if (yr < 1000) {
				printf("WARNING: year %ld in -d option is less than 1000; give all digits\n", yr);
			}
			break;

		case 't':
			/*
			 * To be able to use hms_str2float, we need to convert
			 * colons to h and m.  To avoid overwriting the command
			 * line, make a copy of this argument.
			 */
			strncpy(timebuf, optarg, sizeof (timebuf));
			timebuf[sizeof (timebuf) - 1] = '\0';
			c_ptr = strchr(timebuf, ':');
			if (c_ptr != NULL) {
				*c_ptr = 'h';
				c_ptr = strchr(timebuf, ':');
				if (c_ptr != NULL) {
					*c_ptr = 'm';
				}
			}
			if (hms_str2float(timebuf, &temp) != 0) {
				error("invalid -t option %s", optarg);
			}

			/* need to set the integer breakdown of the time */
			float_to_3_longs(temp, &th, &tm, &ts);
			break;

		case 'z':
			if (sscanf(optarg, "%d", &tz) != 1) {
				error("invalid -z option %s", optarg);
			}
			tzs = "   ";
			break;

		case 'a':
			if (dms_str2float(optarg, &temp) != 0) {
				error("invalid -a option %s", optarg);
			}
			if (temp < -90.0 || temp > 90.0) {
				error("-a option %s must be between -90 and 90 degrees", optarg);
			}
			float_to_3_longs(temp, &latd, &latm, &lats);
			break;

		case 'o':
			if (dms_str2float(optarg, &temp) != 0) {
				error("invalid -o option %s", optarg);
			}
			if (temp < -180.0 || temp > 180.0) {
				error("-o option %s must be between -180 and 180 degrees", optarg);
			}
			float_to_3_longs(temp, &lond, &lonm, &lons);
			break;

		case 'R':
			if (hms_str2float(optarg, &temp) != 0) {
				error("invalid -R option %s", optarg);
			}
			float_to_3_longs(temp, &ra_h, &ra_m, &ra_s);
			break;

		case 'D':
			if (dms_str2float(optarg, &temp) != 0) {
				error("invalid -D option %s", optarg);
			}
			if (temp < -90.0 || temp > 90.0) {
				error("-a option %s must be between -90 and 90 degrees", optarg);
			}
			float_to_3_longs(temp, &dec_d, &dec_m, &dec_s);
			break;

		default:
			usage();
		}
	}

	if (!got['R'] || !got['D']) usage();
}


/*
 * Function:	hms_str2float
 *
 * Description:	This function converts a string representing hours, minutes,
 *		and seconds to a floating point number in hours.  These
 *		formats are allowed:
 *
 * 			HOURS
 * 			HOURS h
 * 			HOURS h MINUTES
 * 			HOURS h MINUTES m
 * 			HOURS h MINUTES m SECONDS
 * 			HOURS h MINUTES m SECONDS s
 *
 *		where HOURS, MINUTES, and SECONDS are numbers having no sign
 *		and there are not actually any blanks between tokens.  Only the
 *		last number given may contain a decimal point.  Whatever
 *		numbers are given must be in these ranges:
 *
 *			0 <= hours < 24
 *			0 <= minutes < 60
 *			0 <= seconds < 60
 *
 *		The sum must be less than 24.
 *
 * Returns:	0 if success, -1 if failure.
 */

int
hms_str2float(

char *str,		/* the string input */
float *ans_ptr		/* answer filled in, valid only on success */
)
{
	char *beg_ptr;		/* point at first char of a number */
	char *end_ptr;		/* point after last char of a number */
	char *c_ptr;		/* point at a char */
	float temp;		/* each number as parsed */


	/* only these characters are allowed in the string */
	if (str[0] == '\0' || strspn(str, "0123456789.hms") != strlen(str)) {
		 return -1;
	}

	/* if a decimal point anywhere, it must be only in the last number */
	c_ptr = strchr(str, '.');
	if (c_ptr != NULL) {
		c_ptr++;
		c_ptr += strspn(c_ptr, "0123456789");
		if (*c_ptr != '\0' && *(c_ptr + 1) != '\0') {
			return -1;
		}
	}

	beg_ptr = str;		/* start of hours number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != 'h' && *end_ptr != '\0') {
		return -1;	/* can't parse hours or junk after it */
	}
	if (temp >= 24.0) {
		return -1;
	}
	*ans_ptr = temp;
	if (*end_ptr == '\0' || *end_ptr == 'h' && *(end_ptr + 1) == '\0') {
		return 0;	/* nothing, or just "h", after hours */
	}

	beg_ptr = end_ptr + 1;	/* start of minutes number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != 'm' && *end_ptr != '\0') {
		return -1;	/* can't parse minutes or junk after it */
	}
	if (temp >= 60.0) {
		 return -1;
	}
	*ans_ptr += temp / 60.0;
	if (*end_ptr == '\0' || *end_ptr == 'm' && *(end_ptr + 1) == '\0') {
		return 0;	/* nothing, or just "m", after minutes */
	}

	beg_ptr = end_ptr + 1;	/* start of seconds number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != 's' && *end_ptr != '\0') {
		return -1;	/* can't parse seconds or junk after it */
	}
	if (temp >= 60.0) {
		return -1;
	}
	*ans_ptr += temp / 3600.0;
	if (*end_ptr == '\0' || *end_ptr == 's' && *(end_ptr + 1) == '\0') {
		return 0;	/* nothing, or just "s", after seconds */
	}

	return -1;		/* extra junk after the end */
}



/*
 * Function:	dms_str2float
 *
 * Description:	This function converts a string representing degrees, minutes,
 *		and seconds to a floating point number in degrees.  These
 *		formats are allowed:
 *
 * 			DEGREES
 * 			DEGREES o
 * 			DEGREES o ARCMINUTES
 * 			DEGREES o ARCMINUTES '
 * 			DEGREES o ARCMINUTES ' ARCSECONDS
 * 			DEGREES o ARCMINUTES ' ARCSECONDS "
 *
 *		where DEGREES, ARCMINUTES, and ARCSECONDS are numbers,
 *		ARCMINUTES and ARCSECONDS having no sign, and DEGREES
 *		optionally having a + or - sign, and there are not actually any
 *		blanks between tokens.  Only the last number given may contain
 *		a decimal point.  If DEGREES has a sign, it is assumed to apply
 *		to all the numbers.  Whatever numbers are given must be in
 *		these ranges (disregarding the sign):
 *
 *			0 <= degrees < 360
 *			0 <= arcminutes < 60
 *			0 <= arcseconds < 60
 *
 *		The sum must be less than 360.
 *
 * Returns:	0 if success, -1 if failure.
 */

int
dms_str2float(

char *str,		/* the string input */
float *ans_ptr		/* answer filled in, valid only on success */
)
{
	char *beg_ptr;		/* point at first char of a number */
	char *end_ptr;		/* point after last char of a number */
	char *c_ptr;		/* point at a char */
	float temp;		/* each number as parsed */
	int sign;		/* -1 if negative input, else 1 */


	/* set sign if present, and skip by it */
	if (str[0] == '-') {
		sign = -1;
		str++;
	} else if (str[0] == '+') {
		sign = 1;
		str++;
	} else {
		sign = 1;
	}

	/* only these characters are allowed in the rest of the string */
	if (str[0] == '\0' || strspn(str, "0123456789.o'\"") != strlen(str)) {
		 return -1;
	}

	/* if a decimal point anywhere, it must be only in the last number */
	c_ptr = strchr(str, '.');
	if (c_ptr != NULL) {
		c_ptr++;
		c_ptr += strspn(c_ptr, "0123456789");
		if (*c_ptr != '\0' && *(c_ptr + 1) != '\0') {
			return -1;
		}
	}

	beg_ptr = str;		/* start of degrees number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != 'o' && *end_ptr != '\0') {
		return -1;	/* can't parse hours or junk after it */
	}
	if (temp > 360.0) {
		 return -1;
	}
	*ans_ptr = temp;
	if (*end_ptr == '\0' || *end_ptr == 'o' && *(end_ptr + 1) == '\0') {
		*ans_ptr *= sign;
		return 0;	/* nothing, or just "o", after degrees */
	}

	beg_ptr = end_ptr + 1;	/* start of arcminutes number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != '\'' && *end_ptr != '\0') {
		return -1;	/* can't parse arcminutes or junk after it */
	}
	if (temp >= 60.0) {
		return -1;
	}
	*ans_ptr += temp / 60.0;
	if (*end_ptr == '\0' || *end_ptr == '\'' && *(end_ptr + 1) == '\0') {
		*ans_ptr *= sign;
		return 0;	/* nothing, or just "'", after arcminutes */
	}

	beg_ptr = end_ptr + 1;	/* start of arcseconds number */
	errno = 0;
	temp = strtof(beg_ptr, &end_ptr);
	if (errno != 0 || end_ptr == beg_ptr ||
			*end_ptr != '"' && *end_ptr != '\0') {
		return -1;	/* can't parse arcseconds or junk after it */
	}
	if (temp >= 60.0) {
		return -1;
	}
	*ans_ptr += temp / 3600.0;
	if (*end_ptr == '\0' || *end_ptr == '"' && *(end_ptr + 1) == '\0') {
		*ans_ptr *= sign;
		return 0;	/* nothing, or just "\"", after arcseconds */
	}

	return -1;
}

void
float_to_3_longs(float f, long *a_ptr, long *b_ptr, long *c_ptr)
{
	int sign;

	/* remember the sign, and make f positive */
	if (f < 0.0) {
		sign = -1;
		f = -f;
	} else {
		sign = 1;
	}

	/* add half a second so that round off is correct */
	f += 0.5 / 3600.0;

	/* break it out into the 3 longs */
	*a_ptr = (long)f;
	*b_ptr = (long)(60 * (f - *a_ptr));
	*c_ptr = (long)(3600 * (f - *a_ptr - *b_ptr / 60.0));

	/* if it was negative, negate all three */
	if (sign == -1) {
		*a_ptr = -*a_ptr;
		*b_ptr = -*b_ptr;
		*c_ptr = -*c_ptr;
	}
}

usage()
{
printf("Usage: %s  -R HhMmSs -D DoM'S\" [-t H:M:S] [-d m/d/y] [-z tz] \\\n\
                [-a DoM'S\"] [-o DoM'S\"]\n\
      where\n\
      R = right ascension (H=hours, M=minutes, S=seconds)\n\
      D = declination (D=degrees, M=arcminutes, S=arcseconds)\n\
      t = time (H=hours, M=mins, S=seconds; 24-hour clock, default=now)\n\
      d = date (month/day/year, must supply all digits of year, default=today)\n\
      z = timezone number (count west from Greenwich, default=central)\n\
      a = latitude (DoM'S\" [D=degrees, M=mins, S=secs] north, default=42o2')\n\
      o = longitude (DoM'S\" [D=degrees, M=mins, S=secs] west, default=88o16')\n\
In all parameters, the (arc)seconds may be omitted, or (arc)minutes and\n\
(arc)seconds may be omitted.  Also, the symbol after the last number (hmso'\":)\n\
may be omitted.  Use negative latitude for south, negative longitude for east.\n\
", Prog);
	exit(1);
}

double aint(z)
	double z;
{
	long trunk;
	
	trunk = (long)z;
	z = (double) trunk;
	return(z);
}

void
praltaz(ra, dec)
double ra;		/* right ascension in radians */
double dec;		/* declination in radians */
{
	double hourangle;	/* in radians; later in hours for printing */
	double altitude;	/* in radians; later in degrees for printing */
	double azimuth;		/* in radians; later in degrees for printing */
	double temp;

	double sinalt;		/* sin(altitude) */
	double t;		/* another temp variable */
	char sign;		/* another temp variable */

	int hr, deg, min, sec;	/* used to set up for printing */

	hourangle = greensid - ra - longitude;
	hourangle = pfmod(hourangle, 2 * PI);
	sinalt = cos(dec) * cos(hourangle) * cos(latitude) +
		sin(dec) * sin(latitude);
	altitude = asin(sinalt);
	temp = -cos(dec) * sin(hourangle) / sqrt(1 - sinalt * sinalt);
	temp = temp >  1.0 ?  1.0 : temp;  /* guard against roundoff error */
	temp = temp < -1.0 ? -1.0 : temp;
	t = acos(temp);
	if (cos(dec) * cos(hourangle) * sin(latitude) <
			sin(dec) * cos(latitude))
		t = -t;
	azimuth = pfmod(t + PI / 2, 2 * PI);

	hourangle *= 12 / PI;		/* convert to hours */
	hourangle += .5 / 3600.;	/* add 1/2 second for rounding */
	if (hourangle >= 24)		/* due to the 1/2 second */
		hourangle -= 24;
	sec = hourangle * 3600;
	hr = sec / 3600;
	min = (sec / 60) % 60;
	sec = sec - 3600 * hr - 60 * min;
	printf("hour angle = %2dh %2dm %2ds", hr, min, sec);

	altitude *= 180 / PI;		/* convert to degrees */
	if (altitude >= 0) {
		sign = '+';
	} else {
		altitude = -altitude;
		sign = '-';
	}
	altitude += .5 / 3600.;		/* add 1/2 second for rounding */
	deg = altitude;
	min = (altitude - deg) * 60;
	sec = (altitude - deg - min / 60.0) * 3600;
	printf("    altitude = %c%2do %2d' %2d\"", sign, deg, min, sec);

	azimuth *= 180 / PI;		/* convert to degrees */
	azimuth += .5 / 3600.;		/* add 1/2 second for rounding */
	if (azimuth >= 360)		/* due to the 1/2 second */
		azimuth -= 360;
	deg = azimuth;
	min = (azimuth - deg) * 60;
	sec = (azimuth - deg - min / 60.0) * 3600;
	printf("   azimuth = %3do %2d' %2d\"", deg, min, sec);
}

/*
 *	pfmod()		find a double modulo another double
 *
 *	This function finds the mathematical modulo of one "double"
 *	number with another.  (The "%" operation in C works only for
 *	integers.)  The second number is assumed to be nonzero, and
 *	the answer will be nonnegative (hence the "p").
 *
 *	Parameters:	a	first number
 *			b	second number
 *
 *	Return value:	a % b
 */

double
pfmod(a, b)

double a;
double b;

{
	long whole;			/* integer part of a / b */
	int reverse;			/* reverse answer if a < 0 */
	double answer;

	if (b < 0.)
		b = -b;

	reverse = 0;
	if (a < 0) {
		a = -a;
		reverse = 1;
	}
	whole = a / b;			/* integer part of division */
	answer = a - b * whole; 	/* fractional part of division */
	if (reverse && answer != 0.)
		answer = b - answer;
	return(answer);
}

/*
 * Function:	error
 *
 * Description:	Print an error message and exit.  It is called with parameters
 *		like printf.  It adds some additional output and prints to
 *		stderr.
 *
 * Returns:	does not return; exits with 1
 */

void
error(char *format, ...)
{
	va_list arglist;


	fprintf(stderr, "%s: error: ", Prog);

	va_start(arglist, format);
	vfprintf(stderr, format, arglist);
	va_end(arglist);

	fprintf(stderr, ".\n");

	exit(1);
}

