/*
 *	precession
 *
 *	This program is given a right ascension and declination for the year
 *	2000, and a year to precess these positions to, and it prints the
 *	precessed right ascension and declination.
 *
 *	Type "precession" with no parameters to see the usage message.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define	PI		3.141592653589793
#define	RAD2DEG(rad)	((rad) * 180.0 / PI)
#define	DEG2RAD(deg)	((deg) * PI / 180.0)
#define	RAD2HR(rad)	((rad) * 12.0 / PI)
#define	HR2RAD(hr)	((hr)  * PI / 12.0)

int
main(int argc, char **argv)
{
	double epsilon_rad;
	double ra_hr;		/* right ascension in hours */
	double dec_deg;		/* declination in degrees */
	double ra_rad;		/* right ascension in radians */
	double dec_rad;		/* declination in radians */
	double x0, y0, z0;	/* starting coords */
	double x1, y1, z1;
	double x2, y2, z2;
	double x3, y3, z3;
	int year;
	int year_interval;
	double p_rad;


	if (argc != 4) {
		fprintf(stderr, "Usage:  %s right_ascension declination year\n",
				argv[0]);
		fprintf(stderr, "	right_ascension and declination are in decimal hours and degrees, respectively\n");
		exit (1);
	}

	ra_hr = atof(argv[1]);
	dec_deg = atof(argv[2]);
	year = atoi(argv[3]);

	year_interval = year - 2000;
	p_rad = (double)year_interval / 25800.0 * 2.0 * PI;

	/* ecliptic obliquity = 23 degrees 26 minutes */
	epsilon_rad = DEG2RAD(23.0 + 26.0 / 60.0);

	/*printf("R.A. = %7.3f hours  Dec = %7.3f degrees\n", ra_hr, dec_deg);*/
	ra_rad = HR2RAD(ra_hr);
	dec_rad = DEG2RAD(dec_deg);

	/* convert R.A and dec to x, y, z */
	x0 = cos(dec_rad) * cos(ra_rad);
	y0 = cos(dec_rad) * sin(ra_rad);
	z0 = sin(dec_rad);

	/*printf("x0 = %7.3f  y0 = %7.3f  z0 = %7.3f\n", x0, y0, z0);*/

	/* rotate clockwise (looking at x axis) to bring ecliptic to equator */
	x1 = x0;
	y1 = y0 * cos(-epsilon_rad) - z0 * sin(-epsilon_rad);
	z1 = y0 * sin(-epsilon_rad) + z0 * cos(-epsilon_rad);

	/*printf("x1 = %7.3f  y1 = %7.3f  z1 = %7.3f\n", x1, y1, z1);*/

	/* rotate clockwise (looking at z axis) to move node along ecliptic */
	x2 = x1 * cos(p_rad) - y1 * sin(p_rad);
	y2 = x1 * sin(p_rad) + y1 * cos(p_rad);
	z2 = z1;

	/*printf("x2 = %7.3f  y2 = %7.3f  z2 = %7.3f\n", x2, y2, z2);*/

	/* rotate counterclockwise (looking at x axis) to re-tilt ecliptic */
	x3 = x2;
	y3 = y2 * cos(epsilon_rad) - z2 * sin(epsilon_rad);
	z3 = y2 * sin(epsilon_rad) + z2 * cos(epsilon_rad);

	/*printf("x3 = %7.3f  y3 = %7.3f  z3 = %7.3f\n", x3, y3, z3);*/

	/* convert x, y, z to R.A. and dec */
	dec_rad = asin(z3);
	if (cos(dec_rad) == 0.0) {
		ra_rad = 0.0;
	} else {
		if (x3 / cos(dec_rad) > 1.0) {
			ra_rad = 0.0;
		} else if (x3 / cos(dec_rad) < -1.0) {
			ra_rad = PI;
		} else {
			ra_rad = acos(x3 / cos(dec_rad));
		}
	}
	if (y3 < 0.0) {
		ra_rad = -ra_rad + 2.0 * PI;
	}

	ra_hr = RAD2HR(ra_rad);
	dec_deg = RAD2DEG(dec_rad);

	printf("R.A. = %7.3f hours  Dec = %7.3f degrees\n", ra_hr, dec_deg);
	exit(0);
}
